/** \file statisticsConservation.cpp
 * \brief File with \ref statisticsConservation class methods implementation.
 *
 * In this file the \ref statisticsConservation class methods are implemented.
 */

#include "utils.h"
#include "statisticsConservation.h"
#include <math.h>
#include <iostream.h>
#include <iomanip.h>


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   statisticsConservation::statisticsConservation()                       |
|       This constructor sets the attributes to 0 or null values           |
|                                                                          |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

statisticsConservation::statisticsConservation(void) {
  
  columns = 0;
  sequences = 0;
  
  halfWindow = 0;
  interv = 0.0;

  Q = NULL;
  MDK = NULL;
  MDK_Pond = NULL;
  MDK_Window = NULL;

  matrixIdentity = NULL;
  simMatrix = NULL;
  sgaps = NULL;   
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   statisticsConservation::~statisticsConservation()                      |
|       Class destructor.                                                  |
|                                                                          |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

statisticsConservation::~statisticsConservation(void) {

  if(Q != NULL)
    memoryDeletion();

}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   statisticsConservation::statisticsConservation(char**, int, int, statisticsGaps) |
|       This constructor calculates some values from the alignment matrix            |
|       and from the statisticsGaps object. It also calculates the row identity      |  
|       matrix.                                                                      |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

statisticsConservation::statisticsConservation(char **alignmentMatrix, int species, int aminos, statisticsGaps *sg){
  
  /* Allocate memory */
  columns = aminos;
  sequences = species;
  sgaps = sg;
  interv = 0.0;
  
  memoryAllocation();
  
  halfWindow = -1;
  
  simMatrix = NULL;

  /* Calculation methods call */
  calculateMatrixIdentity(alignmentMatrix, sequences, columns);  
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   void statisticsConservation::memoryAllocation()                        |
|       Memory allocation for the vectors and attributes of the object.    |
|                                                                          |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsConservation::memoryAllocation(void) {

  int i, j;

  Q = new float[columns];
  MDK = new float[columns];
  MDK_Pond = new float[columns];
  MDK_Window = new float[columns];
  
  for(i = 0; i < columns; i++){
    Q[i] = 0.0;
    MDK[i] = 0.0;
    MDK_Pond[i] = 0.0;
    MDK_Window[i] = 0.0;

  }

  for(i = 0; i < 100; i++)
    freqBlocks[i] = 0;
  
  matrixIdentity = new float*[sequences];
  
  for(i = 0; i < sequences; i++){
    
    matrixIdentity[i] = new float[sequences];
    
    for(j = 0; j < sequences; j++)
      matrixIdentity[i][j] = 0.0;
    
  }
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   void statisticsConservation::memoryDeletion()                          |
|       Memory deletion for the vectors and attributes of the object.      |
|									   |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsConservation::memoryDeletion(void) {

  int i;
  
  delete[] Q;
  delete[] MDK;
  delete[] MDK_Pond;
  delete[] MDK_Window;
  
  for(i = 0; i < sequences; i++)
    delete[] matrixIdentity[i];
  
  delete[] matrixIdentity;
 
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   void statisticsConservation::calculateMatrixIdentity(char **, int, int) |
|        This method calculates the identity function between each pair     |
|        of rows in the alignment, and stores this value in the identity    |
|        matrix.                                                            |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsConservation::calculateMatrixIdentity(char **alignmentMatrix, int species, int aminos){
  
  int i, j, k, sum, length;

  /* For each pair of species */
  for(i = 0; i < species; i++) {
    for(j = i + 1; j < species; j++) {
       
      /* For each position in the alignment of the   */
      /* pair of sequences we are processing           */
      for(k = 0, sum = 0, length = 0; k < aminos; k++) {
	
	/* If we find an aminoacid that is not a gap or an X aminoacid */
	/* in the first sequence of the pair we are processing         */
	if((alignmentMatrix[i][k] != '-') && (alignmentMatrix[i][k] != 'X')) {
	  
	  /* If we also find a valid aminoacid in the second sequence  */
	  if((alignmentMatrix[j][k] != '-') && (alignmentMatrix[j][k] != 'X'))
	    /* If the two valid aminoacids are the same increase the   */
	    /* sum                                                     */
	    if(alignmentMatrix[j][k] ==  alignmentMatrix[i][k]) sum++;
	  /* Increase the length of the sequence free of gaps and X    */
	  /* aminoacids                                                */
	  length++;
	}
	
	/* If the first processed aminoacid is invalid and in the second */
	/* we find a valid aminoacid increase the length of the sequence */
	/* free of gaps and X aminoacids                                 */
	else if((alignmentMatrix[j][k] != '-') && (alignmentMatrix[j][k] != 'X'))
	  length++;
      }
      
      /* Calculate the value of matrixidn for columns j and i */
      matrixIdentity[j][i] = (100.0 - ((float)sum/(float)length) * 100.0);
      matrixIdentity[i][j] = matrixIdentity[j][i];
    }
  }

}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   bool statisticsConservation::calculateVectors(char **, int, int)         |
|        This method calculates statistics vectors from the alignment matrix |
|        and using too the similarity matrix.                                |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

bool statisticsConservation::calculateVectors(char **alignmentMatrix, int species, int aminos) {
  
  int i, j, k;
  float num, den;
  float penal;
  
  /* Get the reference to the gapsWindow and indetermination vector */
  /* used to weight the MDK values                                  */
  const int *numGapsInColumn = sgaps->getGapsWindow();
  const int *indInColumn = sgaps->getIndVector();
  
  /* A similarity matrix must be defined. If not, return false */
  if(simMatrix == NULL) return false;

  /* For each column calculate the Q value, the MD value and the weighted MD value */
  /* using an equation                                                             */
  for(i = 0; i < aminos; i++) {
    for(j = 0, num = 0, den = 0; j < species; j++) {
      if((alignmentMatrix[j][i] != '-') && (alignmentMatrix[j][i] != 'X'))
	for(k = j+1; k < species; k++)
	  if((alignmentMatrix[k][i] != '-') && (alignmentMatrix[k][i] != 'X')) {
	    num += matrixIdentity[j][k] * simMatrix -> getDistance(alignmentMatrix[j][i], alignmentMatrix[k][i]);
	    den += matrixIdentity[j][k];
	  }
    }
    
    /* If we are procesing a column with only one aminoacid, the denominator is 0 */
    /* and we don't execute the division and we set the Q[i] value to 0           */
    if(den == 0)
      Q[i] = 0;
    else
      Q[i] = num / den;
    
    MDK[i] = (float) exp(-Q[i]);
    penal = (1.0 - (float)(numGapsInColumn[i] + indInColumn[i]) / (float)species);
    
    /* Penalize the MDK value with the squared gap penalty */
    MDK_Pond[i] = (float) ((penal * penal) * MDK[i]);
  }

  return true;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   bool statisticsConservation::applyWindow()                            |
|        This method calculates window statistics from the ponderated MDK |
|        values of the MDK_Pond vector.                                   |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

bool statisticsConservation::applyWindow(void) {
  
  int length, i, j, indmax = 0;
  float max = -1.0;
    
  /* If sgaps half window size is the same as the object half window */
  /* stored value, we don't need to recalculate the conservation     */
  /* window vector                                                   */
  if(sgaps->getHalfWindow() == halfWindow)
    return false;

  /* Update the halfWindow value */
  halfWindow = sgaps->getHalfWindow();
  length = 2*halfWindow + 1;
  
  /* Do the average window calculations */
  for(i = 0; i < columns; i++) {
    
    for(j = i - halfWindow; j <= i + halfWindow; j++) {
      
      if(j < 0) 
	MDK_Window[i] += MDK_Pond[-j];
      else if(j >= columns) 
	MDK_Window[i] += MDK_Pond[2*columns - j - 2];
      else  
	MDK_Window[i] += MDK_Pond[j];
      
    }
    
    MDK_Window[i] = MDK_Window[i] / (float)length;
    
    /* Check if the i position value is the new maximum of the alignment */
    if(MDK_Window[i] > max) {
      max = MDK_Window[i];
      indmax = i;
    }
    
  }
  
  /* Calculate the lengtho of the interval. This length will be used to analyze */
  /* frequency values into each interval, since 0 to the maximum conservation   */
  /* value of the alignment (max variable)                                      */
  
  interv = max / 100.0;
  
  for(i = 0; i < indmax; i++)
    freqBlocks[(int)(MDK_Window[i] / interv)]++;
  
  for(i = indmax+1; i < columns; i++)
    freqBlocks[(int)(MDK_Window[i] / interv)]++;
  
  /* Count the maximum in the last interval */
  freqBlocks[99]++;

  return true;
  
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   bool statisticsConservation::setSimilarityMatrix(similarityMatrix)     |
|        This method sets the similarity matrix of the object to the given |
|        matrix.                                                           |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

bool statisticsConservation::setSimilarityMatrix(similarityMatrix *sm) {
  
  if(sm == NULL)
    return false;
  
  simMatrix = sm;
  halfWindow = -1;

  return true;
  
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   float statisticsConservation::calcCutPoint(float)                      |
|        This method returns the cut point value calculated using the two  |
|        given parameters: baseLine(% of columns that sets the security    |
|        minimum when cleaning the alignment) and conservationPct          |
|        (minimum conservation value permitted in the clean alignment.     |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
float statisticsConservation::calcCutPoint(float baseLine, float conservationPct) {
  
  float cutBL;
  
  float *vectAux = new float[columns];
  
  /* Sort a copy of the MDK_Window vector, and take the value of the column */
  /* that marks the % baseline                                              */
  utils::copyVect(MDK_Window, vectAux, columns);
  utils::quicksort(vectAux, 0, columns-1);
  
  cutBL = vectAux[(int) ((float)(columns - 1) * (100.0 - baseLine)/100.0)];

  delete[] vectAux;
  
  /* Return the minimum of the baseLine cut and conservationPct value */
  return (cutBL < conservationPct ? cutBL : conservationPct);
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   float *statisticsConservation::getMdkwVector()                         |
|          This method returns a pointer to the MDK_Window vector          |
|									   |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

float *statisticsConservation::getMdkwVector(void) {

  return MDK_Window;

}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   bool statisticsConservation::isSimMatrixDef()                          |
|          This method checks if a similarity matrix was previously set in |
|          the conservation statistics object.                             |   
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

bool statisticsConservation::isSimMatrixDef(void) {

  return (simMatrix != NULL);

}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|   bool statisticsConservation::vectorsCalculated()                       |
|          This method checks if statistics vectors are calculated         |
|									   |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

bool statisticsConservation::vectorsCalculated(void) {

  return (halfWindow != -1);

}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  void statisticsConservation::printConservationColumns(int)		   |
|	This method prints the conservation value of each column.          |
|									   |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsConservation::printConservationColumns(int aminos) {

  cout.precision(10);
  cout << "Column\t%Conservation" << endl;

  if(MDK_Window != NULL)
    for(int i = 0; i < aminos; i++)
      cout << setiosflags(ios::left) << setw(4) << i << "\t" << setw(7) << MDK_Window[i] << endl;
  else
    for(int i = 0; i < aminos; i++)
      cout << setiosflags(ios::left) << setw(4) << i << "\t" << setw(7) << MDK_Pond[i] << endl;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  void statisticsConservation::printConservationAcl(int)                  |
|	This method prints the accumulated statistics related to the       | 
|       number of columns for a determinated conservation value.           |
|                                                                          |
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsConservation::printConservationAcl(int aminos) {

  int i, num, acm;  
  float refer, *vectAux = new float[aminos];

  utils::copyVect(MDK_Window, vectAux, aminos);
  utils::quicksort(vectAux, 0, aminos-1);

  cout << "N.Col\t%Length\t\tNum\tValue" << endl;
  cout.precision(10);

  refer = vectAux[aminos-1];
  acm = 0; num = 1;

  for(i = aminos-2; i >= 0; i--) {
    acm++;

    if(refer != vectAux[i]) {      
      cout << acm << "\t" << setw(10) << ((float) acm/aminos * 100.0)
           << "\t" << num  << "\t" << setw(15) << refer << endl;
      refer = vectAux[i]; 
      num = 1;
    }
    else num++;
  }
  acm++;  
  cout << acm << "\t" << setw(10) << left << ((float) acm/aminos * 100.0) 
       << "\t" << num  << "\t" << refer << endl;

  delete [] vectAux;
}
